/* ROUTINES COMMON BETWEEN WRITE, AMIN, MESG, ETC */

#include <stdio.h>
#include <string.h>
#include <fcntl.h>
#include <sys/stat.h>
#include "wrttmp.h"
#include "lib_common.h"

extern struct wrthdr wt_head;
extern char *progname;
extern int wstream;


/* INIT_WSTREAM - open the wrttmp file, loading and checking the header.
 * Mode can be either O_RDONLY or O_RDWR.  Prints error message and returns
 * true if it fails.
 */

int init_wstream(char *wrttmp, int mode)
{
    if ((wstream= open(wrttmp,mode)) < 0)
    {
	printf("%s: Unable to open %s to read/write\n", progname, wrttmp);
	return(1);
    }
    fcntl(wstream,F_SETFD,1);		/* Close over execs */

    /* Read in wrttmp file header */
    if (read(wstream,&wt_head,sizeof(struct wrthdr)) == sizeof(struct wrthdr))
    {
	if ( wt_head.hdr_size < sizeof(struct wrthdr) ||
	     wt_head.tmp_size < sizeof(struct wrttmp) )
	{
	    printf("%s: %s file has undersized entries.  Old version?\n",
		progname, wrttmp);
	    return(1);
	}
    }
    else
    {
	/* Initialize header for new wrttmp file */
	wt_head.hdr_size= sizeof(struct wrthdr);
	wt_head.tmp_size= sizeof(struct wrttmp);
	if (mode != O_RDONLY)
	{
	    lseek(wstream, 0L, 0);
	    write(wstream, &wt_head, sizeof(struct wrthdr));
	    lseek(wstream, 0L, 1);	/* must seek between write and read */
	}
    }
    return(0);
}


/* FIND_UTMP -- Find the named tty in the utmp file, returning NULL if we
 * fail.  The tty name need not be null terminated.
 */

struct utmp *find_utmp(char *tty)
{
struct utmp tmputmp;

    strncpy(tmputmp.ut_line, tty, UT_LINESIZE);
    setutent(); /* open and/or rewind */
    return getutline(&tmputmp);
}


/* FIND_WRTTMP:  Find the named tty in the wrttmp file and return the wrttmp
 * field and the offset of that line's entry.  If the entry's login time
 * doesn't match the given one, or if there is no entry in the file, then
 * return a default wrttmp entry instead.  In the latter case, the position
 * returned will be the offset of the first blank slot at the end of the file.
 */

void find_wrttmp(char *tty, time_t time,struct wrttmp *wbuf, long *pos)
{
register int slot;

    slot= 0;
    for(;;)
    {
	lseek(wstream, *pos= wrttmp_offset(slot++),0);
	if (read(wstream, wbuf, sizeof(struct wrttmp)) != sizeof(struct wrttmp))
	    break;

	if (!strncmp(tty, wbuf->wrt_line, UT_LINESIZE))
	{
	    if (wbuf->wrt_time == time)
		return;
	    else
		break;
	}
    }

    /* Set wbuf to default value */
    dflt_wrttmp(wbuf,tty,time);
}


/* DFLT_WRTTMP -- set a wrttmp entry to default status */

void dflt_wrttmp(struct wrttmp *wbuf, char *tty, time_t time)
{
    /* Set wbuf to default value */
    strncpy(wbuf->wrt_line,tty,UT_LINESIZE);
    wbuf->wrt_time= time;
    wbuf->wrt_what[0]= '\0';
    wbuf->wrt_last[0]= '\0';
    wbuf->wrt_pid= -1;
    wbuf->wrt_record= DFLT_RECO;
    wbuf->wrt_telpref= DFLT_PREF;
    wbuf->wrt_modepref= DFLT_MODE;
    wbuf->wrt_bells= DFLT_BELL;
#ifndef TTYPERMS
    wbuf->wrt_mesg= DFLT_MESG;
#endif /*TTYPERMS*/
#ifdef HELPERS
    wbuf->wrt_help= DFLT_HELP;
#endif /*HELPERS*/
#ifdef EXCEPTIONS
    wbuf->wrt_except= DFLT_EXCP;
#endif
}


#ifdef TTYPERMS
int myperms;			/* my original tty perms */
extern char mydevname[];

/* SAVEPERMS:  Remember the current tty permissions of my terminal.  Returns
 * 1 on failure.
 */

int saveperms()
{
struct stat stb;

	if (stat(mydevname,&stb))
	{
		printf("%s: Panic - can't stat %s\n",progname,mydevname);
		return(1);
	}
	myperms= stb.st_mode;
	return(0);
}


/* SETPERMS:  Set the tty permissions to yes (perm='y'), no (perm='n'), or
 * leave them alone (perm='s').
 */

void setperms(char perm)
{
	if (perm == 'n')
		chmod(mydevname,PERMS_OFF);

	if (perm == 'y')
		chmod(mydevname,PERMS_ON);
}


/* RESETPERMS:  Set the tty permissions back to saved state.
 */

void resetperms()
{
	chmod(mydevname,myperms);
}
#endif /*TTYPERMS*/
